@@ -44,39 +44,38 @@
 #endif
 
 /* macros */
-#define MAX(a, b)       ((a) > (b) ? (a) : (b))
-#define MIN(a, b)       ((a) < (b) ? (a) : (b))
-#define BUTTONMASK      (ButtonPressMask|ButtonReleaseMask)
-#define CLEANMASK(mask) (mask & ~(numlockmask|LockMask))
-#define LENGTH(x)       (sizeof x / sizeof x[0])
-#define MAXTAGLEN       16
-#define MOUSEMASK       (BUTTONMASK|PointerMotionMask)
-#define TAGMASK         ((int)((1LL << LENGTH(tags)) - 1))
-#define TEXTW(x)        (textnw(x, strlen(x)) + dc.font.height)
+#define BUTTONMASK              (ButtonPressMask|ButtonReleaseMask)
+#define CLEANMASK(mask)         (mask & ~(numlockmask|LockMask))
+#define INRECT(X,Y,RX,RY,RW,RH) ((X) >= (RX) && (X) < (RX) + (RW) && (Y) >= (RY) && (Y) < (RY) + (RH))
+#define ISVISIBLE(x)            (x->tags & tagset[seltags])
+#define LENGTH(x)               (sizeof x / sizeof x[0])
+#define MAX(a, b)               ((a) > (b) ? (a) : (b))
+#define MIN(a, b)               ((a) < (b) ? (a) : (b))
+#define MAXTAGLEN               16
+#define MOUSEMASK               (BUTTONMASK|PointerMotionMask)
+#define NOBORDER(x)             ((x) - 2 * c->bw)
+#define TAGMASK                 ((int)((1LL << LENGTH(tags)) - 1))
+#define TEXTW(x)                (textnw(x, strlen(x)) + dc.font.height)
 
 /* enums */
 enum { CurNormal, CurResize, CurMove, CurLast };        /* cursor */
 enum { ColBorder, ColFG, ColBG, ColLast };              /* color */
 enum { NetSupported, NetWMName, NetLast };              /* EWMH atoms */
-enum { WMProtocols, WMDelete, WMName, WMState, WMLast };/* default atoms */
-enum { ClkLtSymbol = 64, ClkStatusText, ClkWinTitle,
+enum { WMProtocols, WMDelete, WMState, WMLast };        /* default atoms */
+enum { ClkTagBar, ClkLtSymbol, ClkStatusText, ClkWinTitle,
        ClkClientWin, ClkRootWin, ClkLast };             /* clicks */
 
-/* typedefs */
-typedef unsigned int uint;
-typedef unsigned long ulong;
-
 typedef union {
 	int i;
-	uint ui;
+	unsigned int ui;
 	float f;
 	void *v;
 } Arg;
 
 typedef struct {
-	uint click;
-	uint mask;
-	uint button;
+	unsigned int click;
+	unsigned int mask;
+	unsigned int button;
 	void (*func)(const Arg *arg);
 	const Arg arg;
 } Button;
@@ -88,8 +87,8 @@ struct Client {
 	int x, y, w, h;
 	int basew, baseh, incw, inch, maxw, maxh, minw, minh;
 	int bw, oldbw;
-	uint tags;
-	Bool isbanned, isfixed, isfloating, ismoved, isurgent;
+	unsigned int tags;
+	Bool isfixed, isfloating, isurgent;
 	Client *next;
 	Client *snext;
 	Window win;
@@ -97,8 +96,8 @@ struct Client {
 
 typedef struct {
 	int x, y, w, h;
-	ulong norm[ColLast];
-	ulong sel[ColLast];
+	unsigned long norm[ColLast];
+	unsigned long sel[ColLast];
 	Drawable drawable;
 	GC gc;
 	struct {
@@ -111,7 +110,7 @@ typedef struct {
 } DC; /* draw context */
 
 typedef struct {
-	uint mod;
+	unsigned int mod;
 	KeySym keysym;
 	void (*func)(const Arg *);
 	const Arg arg;
@@ -126,7 +125,7 @@ typedef struct {
 	const char *class;
 	const char *instance;
 	const char *title;
-	uint tags;
+	unsigned int tags;
 	Bool isfloating;
 } Rule;
 
@@ -138,36 +137,36 @@ static void attachstack(Client *c);
 static void buttonpress(XEvent *e);
 static void checkotherwm(void);
 static void cleanup(void);
+static void clearurgent(void);
 static void configure(Client *c);
 static void configurenotify(XEvent *e);
 static void configurerequest(XEvent *e);
 static void destroynotify(XEvent *e);
 static void detach(Client *c);
 static void detachstack(Client *c);
+static void die(const char *errstr, ...);
 static void drawbar(void);
-static void drawsquare(Bool filled, Bool empty, Bool invert, ulong col[ColLast]);
-static void drawtext(const char *text, ulong col[ColLast], Bool invert);
+static void drawsquare(Bool filled, Bool empty, Bool invert, unsigned long col[ColLast]);
+static void drawtext(const char *text, unsigned long col[ColLast], Bool invert);
 static void enternotify(XEvent *e);
-static void eprint(const char *errstr, ...);
 static void expose(XEvent *e);
 static void focus(Client *c);
 static void focusin(XEvent *e);
 static void focusstack(const Arg *arg);
 static Client *getclient(Window w);
-static ulong getcolor(const char *colstr);
+static unsigned long getcolor(const char *colstr);
 static long getstate(Window w);
-static Bool gettextprop(Window w, Atom atom, char *text, uint size);
+static Bool gettextprop(Window w, Atom atom, char *text, unsigned int size);
 static void grabbuttons(Client *c, Bool focused);
 static void grabkeys(void);
 static void initfont(const char *fontstr);
-static Bool isoccupied(uint t);
 static Bool isprotodel(Client *c);
-static Bool isurgent(uint t);
 static void keypress(XEvent *e);
 static void killclient(const Arg *arg);
 static void manage(Window w, XWindowAttributes *wa);
 static void mappingnotify(XEvent *e);
 static void maprequest(XEvent *e);
+static void monocle(void);
 static void movemouse(const Arg *arg);
 static Client *nexttiled(Client *c);
 static void propertynotify(XEvent *e);
@@ -178,22 +177,23 @@ static void restack(void);
 static void run(void);
 static void scan(void);
 static void setclientstate(Client *c, long state);
+static void setlayout(const Arg *arg);
 static void setmfact(const Arg *arg);
 static void setup(void);
+static void showhide(Client *c);
 static void spawn(const Arg *arg);
 static void tag(const Arg *arg);
-static int textnw(const char *text, uint len);
+static int textnw(const char *text, unsigned int len);
 static void tile(void);
 static void togglebar(const Arg *arg);
 static void togglefloating(const Arg *arg);
-static void togglelayout(const Arg *arg);
-static void togglemax(const Arg *arg);
 static void toggletag(const Arg *arg);
 static void toggleview(const Arg *arg);
 static void unmanage(Client *c);
 static void unmapnotify(XEvent *e);
 static void updatebar(void);
 static void updategeom(void);
+static void updatenumlockmask(void);
 static void updatesizehints(Client *c);
 static void updatetitle(Client *c);
 static void updatewmhints(Client *c);
@@ -205,11 +205,13 @@ static void zoom(const Arg *arg);
 
 /* variables */
 static char stext[256];
-static int screen, sx, sy, sw, sh;
-static int by, bh, blw, wx, wy, ww, wh;
-static uint seltags = 0;
+static int screen;
+static int sx, sy, sw, sh; /* X display screen geometry x, y, width, height */ 
+static int by, bh, blw;    /* bar geometry y, height and layout symbol width */
+static int wx, wy, ww, wh; /* window area geometry x, y, width, height, bar excluded */
+static unsigned int seltags = 0, sellt = 0;
 static int (*xerrorxlib)(Display *, XErrorEvent *);
-static uint numlockmask = 0;
+static unsigned int numlockmask = 0;
 static void (*handler[LASTEvent]) (XEvent *) = {
 	[ButtonPress] = buttonpress,
 	[ConfigureRequest] = configurerequest,
@@ -225,72 +227,55 @@ static void (*handler[LASTEvent]) (XEvent *) = {
 	[UnmapNotify] = unmapnotify
 };
 static Atom wmatom[WMLast], netatom[NetLast];
-static Bool ismax = False;
-static Bool otherwm, readin;
+static Bool otherwm;
 static Bool running = True;
-static uint tagset[] = {1, 1}; /* after start, first tag is selected */
 static Client *clients = NULL;
 static Client *sel = NULL;
 static Client *stack = NULL;
 static Cursor cursor[CurLast];
 static Display *dpy;
-static DC dc = {0};
-static Layout *lt = NULL;
+static DC dc;
+static Layout *lt[] = { NULL, NULL };
 static Window root, barwin;
 /* configuration, allows nested code to access above variables */
 #include "config.h"
 
-/* compile-time check if all tags fit into an uint bit array. */
-struct NumTags { char limitexceeded[sizeof(uint) * 8 < LENGTH(tags) ? -1 : 1]; };
+/* compile-time check if all tags fit into an unsigned int bit array. */
+struct NumTags { char limitexceeded[sizeof(unsigned int) * 8 < LENGTH(tags) ? -1 : 1]; };
 
 /* function implementations */
 void
 applyrules(Client *c) {
-	uint i;
+	unsigned int i;
 	Rule *r;
 	XClassHint ch = { 0 };
 
 	/* rule matching */
-	XGetClassHint(dpy, c->win, &ch);
-	for(i = 0; i < LENGTH(rules); i++) {
-		r = &rules[i];
-		if((!r->title || strstr(c->name, r->title))
-		&& (!r->class || (ch.res_class && strstr(ch.res_class, r->class)))
-		&& (!r->instance || (ch.res_name && strstr(ch.res_name, r->instance)))) {
-			c->isfloating = r->isfloating;
-			c->tags |= r->tags & TAGMASK;
+	if(XGetClassHint(dpy, c->win, &ch)) {
+		for(i = 0; i < LENGTH(rules); i++) {
+			r = &rules[i];
+			if((!r->title || strstr(c->name, r->title))
+			&& (!r->class || (ch.res_class && strstr(ch.res_class, r->class)))
+			&& (!r->instance || (ch.res_name && strstr(ch.res_name, r->instance)))) {
+				c->isfloating = r->isfloating;
+				c->tags |= r->tags & TAGMASK;
+			}
 		}
+		if(ch.res_class)
+			XFree(ch.res_class);
+		if(ch.res_name)
+			XFree(ch.res_name);
 	}
-	if(ch.res_class)
-		XFree(ch.res_class);
-	if(ch.res_name)
-		XFree(ch.res_name);
 	if(!c->tags)
 		c->tags = tagset[seltags];
 }
 
 void
 arrange(void) {
-	Client *c;
-
-	for(c = clients; c; c = c->next)
-		if(c->tags & tagset[seltags]) { /* is visible */
-			if(ismax && !c->isfixed) {
-				XMoveResizeWindow(dpy, c->win, wx, wy, ww - 2 * c->bw, wh - 2 * c->bw);
-				c->ismoved = True;
-			}
-			else if(!lt->arrange || c->isfloating)
-				resize(c, c->x, c->y, c->w, c->h, True);
-			c->isbanned = False;
-		}
-		else if(!c->isbanned) {
-			XMoveWindow(dpy, c->win, c->x + 2 * sw, c->y);
-			c->isbanned = c->ismoved = True;
-		}
-
+	showhide(stack);
 	focus(NULL);
-	if(lt->arrange && !ismax)
-		lt->arrange();
+	if(lt[sellt]->arrange)
+		lt[sellt]->arrange();
 	restack();
 }
 
@@ -308,18 +293,19 @@ attachstack(Client *c) {
 
 void
 buttonpress(XEvent *e) {
-	uint i, x, click;
+	unsigned int i, x, click;
+	Arg arg = {0};
 	Client *c;
 	XButtonPressedEvent *ev = &e->xbutton;
 
 	click = ClkRootWin;
 	if(ev->window == barwin) {
 		i = x = 0;
-		do
-			x += TEXTW(tags[i]);
-		while(ev->x >= x && ++i < LENGTH(tags));
-		if(i < LENGTH(tags))
-			click = i;
+		do x += TEXTW(tags[i]); while(ev->x >= x && ++i < LENGTH(tags));
+		if(i < LENGTH(tags)) {
+			click = ClkTagBar;
+			arg.ui = 1 << i;
+		}
 		else if(ev->x < x + blw)
 			click = ClkLtSymbol;
 		else if(ev->x > wx + ww - TEXTW(stext))
@@ -335,32 +321,31 @@ buttonpress(XEvent *e) {
 	for(i = 0; i < LENGTH(buttons); i++)
 		if(click == buttons[i].click && buttons[i].func && buttons[i].button == ev->button
 		   && CLEANMASK(buttons[i].mask) == CLEANMASK(ev->state))
-			buttons[i].func(&buttons[i].arg);
+			buttons[i].func(click == ClkTagBar && buttons[i].arg.i == 0 ? &arg : &buttons[i].arg);
 }
 
 void
 checkotherwm(void) {
 	otherwm = False;
-	XSetErrorHandler(xerrorstart);
+	xerrorxlib = XSetErrorHandler(xerrorstart);
 
 	/* this causes an error if some other window manager is running */
 	XSelectInput(dpy, DefaultRootWindow(dpy), SubstructureRedirectMask);
 	XSync(dpy, False);
 	if(otherwm)
-		eprint("dwm: another window manager is already running\n");
-	XSetErrorHandler(NULL);
-	xerrorxlib = XSetErrorHandler(xerror);
+		die("dwm: another window manager is already running\n");
+	XSetErrorHandler(xerror);
 	XSync(dpy, False);
 }
 
 void
 cleanup(void) {
-	Arg a = {.i = ~0};
+	Arg a = {.ui = ~0};
 	Layout foo = { "", NULL };
 
 	close(STDIN_FILENO);
 	view(&a);
-	lt = &foo;
+	lt[sellt] = &foo;
 	while(stack)
 		unmanage(stack);
 	if(dc.font.set)
@@ -378,6 +363,23 @@ cleanup(void) {
 	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
 }
 
+void
+clearurgent(void) {
+	XWMHints *wmh;
+	Client *c;
+
+	for(c = clients; c; c = c->next)
+		if(ISVISIBLE(c) && c->isurgent) {
+			c->isurgent = False;
+			if (!(wmh = XGetWMHints(dpy, c->win)))
+				continue;
+
+			wmh->flags &= ~XUrgencyHint;
+			XSetWMHints(dpy, c->win, wmh);
+			XFree(wmh);
+		}
+}
+
 void
 configure(Client *c) {
 	XConfigureEvent ce;
@@ -418,9 +420,7 @@ configurerequest(XEvent *e) {
 	if((c = getclient(ev->window))) {
 		if(ev->value_mask & CWBorderWidth)
 			c->bw = ev->border_width;
-		if(ismax && !c->isbanned && !c->isfixed)
-			XMoveResizeWindow(dpy, c->win, wx, wy, ww - 2 * c->bw, wh + 2 * c->bw);
-		else if(c->isfloating || !lt->arrange) {
+		else if(c->isfloating || !lt[sellt]->arrange) {
 			if(ev->value_mask & CWX)
 				c->x = sx + ev->x;
 			if(ev->value_mask & CWY)
@@ -433,10 +433,9 @@ configurerequest(XEvent *e) {
 				c->x = sx + (sw / 2 - c->w / 2); /* center in x direction */
 			if((c->y - sy + c->h) > sh && c->isfloating)
 				c->y = sy + (sh / 2 - c->h / 2); /* center in y direction */
-			if((ev->value_mask & (CWX|CWY))
-			&& !(ev->value_mask & (CWWidth|CWHeight)))
+			if((ev->value_mask & (CWX|CWY)) && !(ev->value_mask & (CWWidth|CWHeight)))
 				configure(c);
-			if(!c->isbanned)
+			if(ISVISIBLE(c))
 				XMoveResizeWindow(dpy, c->win, c->x, c->y, c->w, c->h);
 		}
 		else
@@ -466,16 +465,10 @@ destroynotify(XEvent *e) {
 
 void
 detach(Client *c) {
-	Client *i;
+	Client **tc;
 
-	if (c != clients) {
-		for(i = clients; i->next != c; i = i->next);
-		i->next = c->next;
-	}
-	else {
-		clients = c->next;
-	}
-	c->next =  NULL;
+	for(tc = &clients; *tc && *tc != c; tc = &(*tc)->next);
+	*tc = c->next;
 }
 
 void
@@ -486,28 +479,40 @@ detachstack(Client *c) {
 	*tc = c->snext;
 }
 
+void
+die(const char *errstr, ...) {
+	va_list ap;
+
+	va_start(ap, errstr);
+	vfprintf(stderr, errstr, ap);
+	va_end(ap);
+	exit(EXIT_FAILURE);
+}
+
 void
 drawbar(void) {
-	int i, x;
+	int x;
+	unsigned int i, occ = 0, urg = 0;
+	unsigned long *col;
 	Client *c;
 
+	for(c = clients; c; c = c->next) {
+		occ |= c->tags;
+		if(c->isurgent)
+			urg |= c->tags;
+	}
+
 	dc.x = 0;
-	for(c = stack; c && c->isbanned; c = c->snext);
 	for(i = 0; i < LENGTH(tags); i++) {
 		dc.w = TEXTW(tags[i]);
-		if(tagset[seltags] & 1 << i) {
-			drawtext(tags[i], dc.sel, isurgent(i));
-			drawsquare(c && c->tags & 1 << i, isoccupied(i), isurgent(i), dc.sel);
-		}
-		else {
-			drawtext(tags[i], dc.norm, isurgent(i));
-			drawsquare(c && c->tags & 1 << i, isoccupied(i), isurgent(i), dc.norm);
-		}
+		col = tagset[seltags] & 1 << i ? dc.sel : dc.norm;
+		drawtext(tags[i], col, urg & 1 << i);
+		drawsquare(sel && sel->tags & 1 << i, occ & 1 << i, urg & 1 << i, col);
 		dc.x += dc.w;
 	}
 	if(blw > 0) {
 		dc.w = blw;
-		drawtext(lt->symbol, dc.norm, ismax);
+		drawtext(lt[sellt]->symbol, dc.norm, False);
 		x = dc.x + dc.w;
 	}
 	else
@@ -521,9 +526,9 @@ drawbar(void) {
 	drawtext(stext, dc.norm, False);
 	if((dc.w = dc.x - x) > bh) {
 		dc.x = x;
-		if(c) {
-			drawtext(c->name, dc.sel, False);
-			drawsquare(c->isfixed, c->isfloating, False, dc.sel);
+		if(sel) {
+			drawtext(sel->name, dc.sel, False);
+			drawsquare(sel->isfixed, sel->isfloating, False, dc.sel);
 		}
 		else
 			drawtext(NULL, dc.norm, False);
@@ -533,7 +538,7 @@ drawbar(void) {
 }
 
 void
-drawsquare(Bool filled, Bool empty, Bool invert, ulong col[ColLast]) {
+drawsquare(Bool filled, Bool empty, Bool invert, unsigned long col[ColLast]) {
 	int x;
 	XGCValues gcv;
 	XRectangle r = { dc.x, dc.y, dc.w, dc.h };
@@ -554,25 +559,24 @@ drawsquare(Bool filled, Bool empty, Bool invert, ulong col[ColLast]) {
 }
 
 void
-drawtext(const char *text, ulong col[ColLast], Bool invert) {
+drawtext(const char *text, unsigned long col[ColLast], Bool invert) {
+	char buf[256];
 	int i, x, y, h, len, olen;
 	XRectangle r = { dc.x, dc.y, dc.w, dc.h };
-	char buf[256];
 
 	XSetForeground(dpy, dc.gc, col[invert ? ColFG : ColBG]);
 	XFillRectangles(dpy, dc.drawable, dc.gc, &r, 1);
 	if(!text)
 		return;
 	olen = strlen(text);
-	len = MIN(olen, sizeof buf);
-	memcpy(buf, text, len);
 	h = dc.font.ascent + dc.font.descent;
 	y = dc.y + (dc.h / 2) - (h / 2) + dc.font.ascent;
 	x = dc.x + (h / 2);
 	/* shorten text if necessary */
-	for(; len && (i = textnw(buf, len)) > dc.w - h; len--);
+	for(len = MIN(olen, sizeof buf); len && textnw(text, len) > dc.w - h; len--);
 	if(!len)
 		return;
+	memcpy(buf, text, len);
 	if(len < olen)
 		for(i = len; i && i > len - 3; buf[--i] = '.');
 	XSetForeground(dpy, dc.gc, col[invert ? ColBG : ColFG]);
@@ -595,16 +599,6 @@ enternotify(XEvent *e) {
 		focus(NULL);
 }
 
-void
-eprint(const char *errstr, ...) {
-	va_list ap;
-
-	va_start(ap, errstr);
-	vfprintf(stderr, errstr, ap);
-	va_end(ap);
-	exit(EXIT_FAILURE);
-}
-
 void
 expose(XEvent *e) {
 	XExposeEvent *ev = &e->xexpose;
@@ -615,8 +609,8 @@ expose(XEvent *e) {
 
 void
 focus(Client *c) {
-	if(!c || c->isbanned)
-		for(c = stack; c && c->isbanned; c = c->snext);
+	if(!c || !ISVISIBLE(c))
+		for(c = stack; c && !ISVISIBLE(c); c = c->snext);
 	if(sel && sel != c) {
 		grabbuttons(sel, False);
 		XSetWindowBorder(dpy, sel->win, dc.norm[ColBorder]);
@@ -649,17 +643,17 @@ focusstack(const Arg *arg) {
 	if(!sel)
 		return;
 	if (arg->i > 0) {
-		for(c = sel->next; c && c->isbanned; c = c->next);
+		for(c = sel->next; c && !ISVISIBLE(c); c = c->next);
 		if(!c)
-			for(c = clients; c && c->isbanned; c = c->next);
+			for(c = clients; c && !ISVISIBLE(c); c = c->next);
 	}
 	else {
 		for(i = clients; i != sel; i = i->next)
-			if (!i->isbanned)
+			if(ISVISIBLE(i))
 				c = i;
 		if(!c)
 			for(; i; i = i->next)
-				if (!i->isbanned)
+				if(ISVISIBLE(i))
 					c = i;
 	}
 	if(c) {
@@ -676,13 +670,13 @@ getclient(Window w) {
 	return c;
 }
 
-ulong
+unsigned long
 getcolor(const char *colstr) {
 	Colormap cmap = DefaultColormap(dpy, screen);
 	XColor color;
 
 	if(!XAllocNamedColor(dpy, cmap, colstr, &color, &color))
-		eprint("error, cannot allocate color '%s'\n", colstr);
+		die("error, cannot allocate color '%s'\n", colstr);
 	return color.pixel;
 }
 
@@ -691,7 +685,7 @@ getstate(Window w) {
 	int format, status;
 	long result = -1;
 	unsigned char *p = NULL;
-	ulong n, extra;
+	unsigned long n, extra;
 	Atom real;
 
 	status = XGetWindowProperty(dpy, w, wmatom[WMState], 0L, 2L, False, wmatom[WMState],
@@ -705,7 +699,7 @@ getstate(Window w) {
 }
 
 Bool
-gettextprop(Window w, Atom atom, char *text, uint size) {
+gettextprop(Window w, Atom atom, char *text, unsigned int size) {
 	char **list = NULL;
 	int n;
 	XTextProperty name;
@@ -732,47 +726,37 @@ gettextprop(Window w, Atom atom, char *text, uint size) {
 
 void
 grabbuttons(Client *c, Bool focused) {
-	int i, j;
-	uint buttons[]   = { Button1, Button2, Button3 };
-	uint modifiers[] = { MODKEY, MODKEY|LockMask, MODKEY|numlockmask, MODKEY|numlockmask|LockMask };
-
-	XUngrabButton(dpy, AnyButton, AnyModifier, c->win);
-	if(focused)
-		for(i = 0; i < LENGTH(buttons); i++)
-			for(j = 0; j < LENGTH(modifiers); j++)
-				XGrabButton(dpy, buttons[i], modifiers[j], c->win, False,
-					BUTTONMASK, GrabModeAsync, GrabModeSync, None, None);
-	else
-		XGrabButton(dpy, AnyButton, AnyModifier, c->win, False,
-			BUTTONMASK, GrabModeAsync, GrabModeSync, None, None);
+	updatenumlockmask();
+	{
+		unsigned int i, j;
+		unsigned int modifiers[] = { 0, LockMask, numlockmask, numlockmask|LockMask };
+		XUngrabButton(dpy, AnyButton, AnyModifier, c->win);
+		if(focused) {
+			for(i = 0; i < LENGTH(buttons); i++)
+				if(buttons[i].click == ClkClientWin)
+					for(j = 0; j < LENGTH(modifiers); j++)
+						XGrabButton(dpy, buttons[i].button, buttons[i].mask | modifiers[j], c->win, False, BUTTONMASK, GrabModeAsync, GrabModeSync, None, None);
+		} else
+			XGrabButton(dpy, AnyButton, AnyModifier, c->win, False,
+			            BUTTONMASK, GrabModeAsync, GrabModeSync, None, None);
+	}
 }
 
 void
 grabkeys(void) {
-	uint i, j;
-	KeyCode code;
-	XModifierKeymap *modmap;
+	updatenumlockmask();
+	{ /* grab keys */
+		unsigned int i, j;
+		unsigned int modifiers[] = { 0, LockMask, numlockmask, numlockmask|LockMask };
+		KeyCode code;
 
-	/* init modifier map */
-	modmap = XGetModifierMapping(dpy);
-	for(i = 0; i < 8; i++)
-		for(j = 0; j < modmap->max_keypermod; j++) {
-			if(modmap->modifiermap[i * modmap->max_keypermod + j] == XKeysymToKeycode(dpy, XK_Num_Lock))
-				numlockmask = (1 << i);
+		XUngrabKey(dpy, AnyKey, AnyModifier, root);
+		for(i = 0; i < LENGTH(keys); i++) {
+			if((code = XKeysymToKeycode(dpy, keys[i].keysym)))
+				for(j = 0; j < LENGTH(modifiers); j++)
+					XGrabKey(dpy, code, keys[i].mod | modifiers[j], root,
+						 True, GrabModeAsync, GrabModeAsync);
 		}
-	XFreeModifiermap(modmap);
-
-	XUngrabKey(dpy, AnyKey, AnyModifier, root);
-	for(i = 0; i < LENGTH(keys); i++) {
-		code = XKeysymToKeycode(dpy, keys[i].keysym);
-		XGrabKey(dpy, code, keys[i].mod, root, True,
-				GrabModeAsync, GrabModeAsync);
-		XGrabKey(dpy, code, keys[i].mod|LockMask, root, True,
-				GrabModeAsync, GrabModeAsync);
-		XGrabKey(dpy, code, keys[i].mod|numlockmask, root, True,
-				GrabModeAsync, GrabModeAsync);
-		XGrabKey(dpy, code, keys[i].mod|numlockmask|LockMask, root, True,
-				GrabModeAsync, GrabModeAsync);
 	}
 }
 
@@ -782,8 +766,6 @@ initfont(const char *fontstr) {
 	int i, n;
 
 	missing = NULL;
-	if(dc.font.set)
-		XFreeFontSet(dpy, dc.font.set);
 	dc.font.set = XCreateFontSet(dpy, fontstr, &missing, &n, &def);
 	if(missing) {
 		while(n--)
@@ -804,28 +786,15 @@ initfont(const char *fontstr) {
 		}
 	}
 	else {
-		if(dc.font.xfont)
-			XFreeFont(dpy, dc.font.xfont);
-		dc.font.xfont = NULL;
 		if(!(dc.font.xfont = XLoadQueryFont(dpy, fontstr))
 		&& !(dc.font.xfont = XLoadQueryFont(dpy, "fixed")))
-			eprint("error, cannot load font: '%s'\n", fontstr);
+			die("error, cannot load font: '%s'\n", fontstr);
 		dc.font.ascent = dc.font.xfont->ascent;
 		dc.font.descent = dc.font.xfont->descent;
 	}
 	dc.font.height = dc.font.ascent + dc.font.descent;
 }
 
-Bool
-isoccupied(uint t) {
-	Client *c;
-
-	for(c = clients; c; c = c->next)
-		if(c->tags & 1 << t)
-			return True;
-	return False;
-}
-
 Bool
 isprotodel(Client *c) {
 	int i, n;
@@ -841,19 +810,9 @@ isprotodel(Client *c) {
 	return ret;
 }
 
-Bool
-isurgent(uint t) {
-	Client *c;
-
-	for(c = clients; c; c = c->next)
-		if(c->isurgent && c->tags & 1 << t)
-			return True;
-	return False;
-}
-
 void
 keypress(XEvent *e) {
-	uint i;
+	unsigned int i;
 	KeySym keysym;
 	XKeyEvent *ev;
 
@@ -887,13 +846,14 @@ killclient(const Arg *arg) {
 
 void
 manage(Window w, XWindowAttributes *wa) {
+	static Client cz;
 	Client *c, *t = NULL;
-	Status rettrans;
-	Window trans;
+	Window trans = None;
 	XWindowChanges wc;
 
-	if(!(c = calloc(1, sizeof(Client))))
-		eprint("fatal: could not calloc() %u bytes\n", sizeof(Client));
+	if(!(c = malloc(sizeof(Client))))
+		die("fatal: could not malloc() %u bytes\n", sizeof(Client));
+	*c = cz;
 	c->win = w;
 
 	/* geometry */
@@ -905,13 +865,13 @@ manage(Window w, XWindowAttributes *wa) {
 	if(c->w == sw && c->h == sh) {
 		c->x = sx;
 		c->y = sy;
-		c->bw = wa->border_width;
+		c->bw = 0;
 	}
 	else {
 		if(c->x + c->w + 2 * c->bw > sx + sw)
-			c->x = sx + sw - c->w - 2 * c->bw;
+			c->x = sx + sw - NOBORDER(c->w);
 		if(c->y + c->h + 2 * c->bw > sy + sh)
-			c->y = sy + sh - c->h - 2 * c->bw;
+			c->y = sy + sh - NOBORDER(c->h);
 		c->x = MAX(c->x, sx);
 		/* only fix client y-offset, if the client center might cover the bar */
 		c->y = MAX(c->y, ((by == 0) && (c->x + (c->w / 2) >= wx) && (c->x + (c->w / 2) < wx + ww)) ? bh : sy);
@@ -926,19 +886,19 @@ manage(Window w, XWindowAttributes *wa) {
 	XSelectInput(dpy, w, EnterWindowMask|FocusChangeMask|PropertyChangeMask|StructureNotifyMask);
 	grabbuttons(c, False);
 	updatetitle(c);
-	if((rettrans = XGetTransientForHint(dpy, w, &trans) == Success))
-		for(t = clients; t && t->win != trans; t = t->next);
+	if(XGetTransientForHint(dpy, w, &trans))
+		t = getclient(trans);
 	if(t)
 		c->tags = t->tags;
 	else
 		applyrules(c);
 	if(!c->isfloating)
-		c->isfloating = (rettrans == Success) || c->isfixed;
+		c->isfloating = trans != None || c->isfixed;
 	if(c->isfloating)
 		XRaiseWindow(dpy, c->win);
 	attach(c);
 	attachstack(c);
-	XMoveResizeWindow(dpy, c->win, c->x, c->y, c->w, c->h); /* some windows require this */
+	XMoveResizeWindow(dpy, c->win, c->x + 2 * sw, c->y, c->w, c->h); /* some windows require this */
 	XMapWindow(dpy, c->win);
 	setclientstate(c, NormalState);
 	arrange();
@@ -966,10 +926,18 @@ maprequest(XEvent *e) {
 		manage(ev->window, &wa);
 }
 
+void
+monocle(void) {
+	Client *c;
+
+	for(c = nexttiled(clients); c; c = nexttiled(c->next))
+		resize(c, wx, wy, NOBORDER(ww), NOBORDER(wh), resizehints);
+}
+
 void
 movemouse(const Arg *arg) {
-	int x1, y1, ocx, ocy, di, nx, ny;
-	uint dui;
+	int x, y, ocx, ocy, di, nx, ny;
+	unsigned int dui;
 	Client *c;
 	Window dummy;
 	XEvent ev;
@@ -977,18 +945,15 @@ movemouse(const Arg *arg) {
 	if(!(c = sel))
 		return;
 	restack();
-	ocx = nx = c->x;
-	ocy = ny = c->y;
+	ocx = c->x;
+	ocy = c->y;
 	if(XGrabPointer(dpy, root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync,
 	None, cursor[CurMove], CurrentTime) != GrabSuccess)
 		return;
-	XQueryPointer(dpy, root, &dummy, &dummy, &x1, &y1, &di, &di, &dui);
-	for(;;) {
+	XQueryPointer(dpy, root, &dummy, &dummy, &x, &y, &di, &di, &dui);
+	do {
 		XMaskEvent(dpy, MOUSEMASK|ExposureMask|SubstructureRedirectMask, &ev);
 		switch (ev.type) {
-		case ButtonRelease:
-			XUngrabPointer(dpy, CurrentTime);
-			return;
 		case ConfigureRequest:
 		case Expose:
 		case MapRequest:
@@ -996,31 +961,33 @@ movemouse(const Arg *arg) {
 			break;
 		case MotionNotify:
 			XSync(dpy, False);
-			nx = ocx + (ev.xmotion.x - x1);
-			ny = ocy + (ev.xmotion.y - y1);
+			nx = ocx + (ev.xmotion.x - x);
+			ny = ocy + (ev.xmotion.y - y);
 			if(snap && nx >= wx && nx <= wx + ww
 			        && ny >= wy && ny <= wy + wh) {
 				if(abs(wx - nx) < snap)
 					nx = wx;
 				else if(abs((wx + ww) - (nx + c->w + 2 * c->bw)) < snap)
-					nx = wx + ww - c->w - 2 * c->bw;
+					nx = wx + ww - NOBORDER(c->w);
 				if(abs(wy - ny) < snap)
 					ny = wy;
 				else if(abs((wy + wh) - (ny + c->h + 2 * c->bw)) < snap)
-					ny = wy + wh - c->h - 2 * c->bw;
-				if(!c->isfloating && lt->arrange && (abs(nx - c->x) > snap || abs(ny - c->y) > snap))
+					ny = wy + wh - NOBORDER(c->h);
+				if(!c->isfloating && lt[sellt]->arrange && (abs(nx - c->x) > snap || abs(ny - c->y) > snap))
 					togglefloating(NULL);
 			}
-			if(!lt->arrange || c->isfloating)
+			if(!lt[sellt]->arrange || c->isfloating)
 				resize(c, nx, ny, c->w, c->h, False);
 			break;
 		}
 	}
+	while(ev.type != ButtonRelease);
+	XUngrabPointer(dpy, CurrentTime);
 }
 
 Client *
 nexttiled(Client *c) {
-	for(; c && (c->isfloating || c->isbanned); c = c->next);
+	for(; c && (c->isfloating || !ISVISIBLE(c)); c = c->next);
 	return c;
 }
 
@@ -1066,22 +1033,31 @@ resize(Client *c, int x, int y, int w, int h, Bool sizehints) {
 	XWindowChanges wc;
 
 	if(sizehints) {
+		/* see last two sentences in ICCCM 4.1.2.3 */
+		Bool baseismin = c->basew == c->minw && c->baseh == c->minh;
+
 		/* set minimum possible */
 		w = MAX(1, w);
 		h = MAX(1, h);
 
-		/* temporarily remove base dimensions */
-		w -= c->basew;
-		h -= c->baseh;
+		if(!baseismin) { /* temporarily remove base dimensions */
+			w -= c->basew;
+			h -= c->baseh;
+		}
 
 		/* adjust for aspect limits */
 		if(c->mina > 0 && c->maxa > 0) {
-			if(c->maxa < (float) w/h)
+			if(c->maxa < (float)w / h)
 				w = h * c->maxa;
-			else if(c->mina > (float) h/w)
+			else if(c->mina < (float)h / w)
 				h = w * c->mina;
 		}
 
+		if(baseismin) { /* increment calculation requires this */
+			w -= c->basew;
+			h -= c->baseh;
+		}
+
 		/* adjust for increment value */
 		if(c->incw)
 			w -= w % c->incw;
@@ -1104,9 +1080,9 @@ resize(Client *c, int x, int y, int w, int h, Bool sizehints) {
 	if(w <= 0 || h <= 0)
 		return;
 	if(x > sx + sw)
-		x = sw - w - 2 * c->bw;
+		x = sw - NOBORDER(w);
 	if(y > sy + sh)
-		y = sh - h - 2 * c->bw;
+		y = sh - NOBORDER(h);
 	if(x + w + 2 * c->bw < sx)
 		x = sx;
 	if(y + h + 2 * c->bw < sy)
@@ -1115,8 +1091,7 @@ resize(Client *c, int x, int y, int w, int h, Bool sizehints) {
 		h = bh;
 	if(w < bh)
 		w = bh;
-	if(c->x != x || c->y != y || c->w != w || c->h != h || c->ismoved) {
-		c->ismoved = False;
+	if(c->x != x || c->y != y || c->w != w || c->h != h) {
 		c->x = wc.x = x;
 		c->y = wc.y = y;
 		c->w = wc.width = w;
@@ -1145,15 +1120,9 @@ resizemouse(const Arg *arg) {
 	None, cursor[CurResize], CurrentTime) != GrabSuccess)
 		return;
 	XWarpPointer(dpy, None, c->win, 0, 0, 0, 0, c->w + c->bw - 1, c->h + c->bw - 1);
-	for(;;) {
-		XMaskEvent(dpy, MOUSEMASK|ExposureMask|SubstructureRedirectMask , &ev);
+	do {
+		XMaskEvent(dpy, MOUSEMASK|ExposureMask|SubstructureRedirectMask, &ev);
 		switch(ev.type) {
-		case ButtonRelease:
-			XWarpPointer(dpy, None, c->win, 0, 0, 0, 0,
-					c->w + c->bw - 1, c->h + c->bw - 1);
-			XUngrabPointer(dpy, CurrentTime);
-			while(XCheckMaskEvent(dpy, EnterWindowMask, &ev));
-			return;
 		case ConfigureRequest:
 		case Expose:
 		case MapRequest:
@@ -1161,20 +1130,24 @@ resizemouse(const Arg *arg) {
 			break;
 		case MotionNotify:
 			XSync(dpy, False);
-			nw = MAX(ev.xmotion.x - ocx - 2 * c->bw + 1, 1);
-			nh = MAX(ev.xmotion.y - ocy - 2 * c->bw + 1, 1);
+			nw = MAX(ev.xmotion.x - NOBORDER(ocx) + 1, 1);
+			nh = MAX(ev.xmotion.y - NOBORDER(ocy) + 1, 1);
 
 			if(snap && nw >= wx && nw <= wx + ww
 			        && nh >= wy && nh <= wy + wh) {
-				if(!c->isfloating && lt->arrange
+				if(!c->isfloating && lt[sellt]->arrange
 				   && (abs(nw - c->w) > snap || abs(nh - c->h) > snap))
 					togglefloating(NULL);
 			}
-			if(!lt->arrange || c->isfloating)
+			if(!lt[sellt]->arrange || c->isfloating)
 				resize(c, c->x, c->y, nw, nh, True);
 			break;
 		}
 	}
+	while(ev.type != ButtonRelease);
+	XWarpPointer(dpy, None, c->win, 0, 0, 0, 0, c->w + c->bw - 1, c->h + c->bw - 1);
+	XUngrabPointer(dpy, CurrentTime);
+	while(XCheckMaskEvent(dpy, EnterWindowMask, &ev));
 }
 
 void
@@ -1186,13 +1159,13 @@ restack(void) {
 	drawbar();
 	if(!sel)
 		return;
-	if(ismax || sel->isfloating || !lt->arrange)
+	if(sel->isfloating || !lt[sellt]->arrange)
 		XRaiseWindow(dpy, sel->win);
-	if(!ismax && lt->arrange) {
+	if(lt[sellt]->arrange) {
 		wc.stack_mode = Below;
 		wc.sibling = barwin;
 		for(c = stack; c; c = c->snext)
-			if(!c->isfloating && !c->isbanned) {
+			if(!c->isfloating && ISVISIBLE(c)) {
 				XConfigureWindow(dpy, c->win, CWSibling|CWStackMode, &wc);
 				wc.sibling = c->win;
 			}
@@ -1207,13 +1180,12 @@ run(void) {
 	char sbuf[sizeof stext];
 	fd_set rd;
 	int r, xfd;
-	uint len, offset;
+	unsigned int len, offset;
 	XEvent ev;
 
 	/* main event loop, also reads status text from stdin */
 	XSync(dpy, False);
 	xfd = ConnectionNumber(dpy);
-	readin = True;
 	offset = 0;
 	len = sizeof stext - 1;
 	sbuf[len] = stext[len] = '\0'; /* 0-terminator is never touched */
@@ -1225,7 +1197,7 @@ run(void) {
 		if(select(xfd + 1, &rd, NULL, NULL, NULL) == -1) {
 			if(errno == EINTR)
 				continue;
-			eprint("select failed\n");
+			die("select failed\n");
 		}
 		if(FD_ISSET(STDIN_FILENO, &rd)) {
 			switch((r = read(STDIN_FILENO, sbuf + offset, len - offset))) {
@@ -1263,11 +1235,10 @@ run(void) {
 
 void
 scan(void) {
-	uint i, num;
-	Window *wins, d1, d2;
+	unsigned int i, num;
+	Window d1, d2, *wins = NULL;
 	XWindowAttributes wa;
 
-	wins = NULL;
 	if(XQueryTree(dpy, root, &d1, &d2, &wins, &num)) {
 		for(i = 0; i < num; i++) {
 			if(!XGetWindowAttributes(dpy, wins[i], &wa)
@@ -1283,9 +1254,9 @@ scan(void) {
 			&& (wa.map_state == IsViewable || getstate(wins[i]) == IconicState))
 				manage(wins[i], &wa);
 		}
+		if(wins)
+			XFree(wins);
 	}
-	if(wins)
-		XFree(wins);
 }
 
 void
@@ -1296,12 +1267,24 @@ setclientstate(Client *c, long state) {
 			PropModeReplace, (unsigned char *)data, 2);
 }
 
+void
+setlayout(const Arg *arg) {
+	if(!arg || !arg->v || arg->v != lt[sellt])
+		sellt ^= 1;
+	if(arg && arg->v)
+		lt[sellt] = (Layout *)arg->v;
+	if(sel)
+		arrange();
+	else
+		drawbar();
+}
+
 /* arg > 1.0 will set mfact absolutly */
 void
 setmfact(const Arg *arg) {
 	float f;
 
-	if(!arg || !lt->arrange)
+	if(!arg || !lt[sellt]->arrange)
 		return;
 	f = arg->f < 1.0 ? arg->f + mfact : arg->f - 1.0;
 	if(f < 0.1 || f > 0.9)
@@ -1312,7 +1295,7 @@ setmfact(const Arg *arg) {
 
 void
 setup(void) {
-	uint i;
+	unsigned int i;
 	int w;
 	XSetWindowAttributes wa;
 
@@ -1325,13 +1308,13 @@ setup(void) {
 	sw = DisplayWidth(dpy, screen);
 	sh = DisplayHeight(dpy, screen);
 	bh = dc.h = dc.font.height + 2;
-	lt = layouts;
+	lt[0] = &layouts[0];
+	lt[1] = &layouts[1 % LENGTH(layouts)];
 	updategeom();
 
 	/* init atoms */
 	wmatom[WMProtocols] = XInternAtom(dpy, "WM_PROTOCOLS", False);
 	wmatom[WMDelete] = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
-	wmatom[WMName] = XInternAtom(dpy, "WM_NAME", False);
 	wmatom[WMState] = XInternAtom(dpy, "WM_STATE", False);
 	netatom[NetSupported] = XInternAtom(dpy, "_NET_SUPPORTED", False);
 	netatom[NetWMName] = XInternAtom(dpy, "_NET_WM_NAME", False);
@@ -1382,11 +1365,25 @@ setup(void) {
 	XChangeWindowAttributes(dpy, root, CWEventMask|CWCursor, &wa);
 	XSelectInput(dpy, root, wa.event_mask);
 
-
-	/* grab keys */
 	grabkeys();
 }
 
+void
+showhide(Client *c) {
+	if(!c)
+		return;
+	if(ISVISIBLE(c)) { /* show clients top down */
+		XMoveWindow(dpy, c->win, c->x, c->y);
+		if(!lt[sellt]->arrange || c->isfloating)
+			resize(c, c->x, c->y, c->w, c->h, True);
+		showhide(c->snext);
+	}
+	else { /* hide clients bottom up */
+		showhide(c->snext);
+		XMoveWindow(dpy, c->win, c->x + 2 * sw, c->y);
+	}
+}
+
 void
 spawn(const Arg *arg) {
 	/* The double-fork construct avoids zombie processes and keeps the code
@@ -1414,7 +1411,7 @@ tag(const Arg *arg) {
 }
 
 int
-textnw(const char *text, uint len) {
+textnw(const char *text, unsigned int len) {
 	XRectangle r;
 
 	if(dc.font.set) {
@@ -1427,7 +1424,7 @@ textnw(const char *text, uint len) {
 void
 tile(void) {
 	int x, y, h, w, mw;
-	uint i, n;
+	unsigned int i, n;
 	Client *c;
 
 	for(n = 0, c = nexttiled(clients); c; c = nexttiled(c->next), n++);
@@ -1437,7 +1434,7 @@ tile(void) {
 	/* master */
 	c = nexttiled(clients);
 	mw = mfact * ww;
-	resize(c, wx, wy, (n == 1 ? ww : mw) - 2 * c->bw, wh - 2 * c->bw, resizehints);
+	resize(c, wx, wy, NOBORDER(n == 1 ? ww : mw), NOBORDER(wh), resizehints);
 
 	if(--n == 0)
 		return;
@@ -1451,8 +1448,8 @@ tile(void) {
 		h = wh;
 
 	for(i = 0, c = nexttiled(c->next); c; c = nexttiled(c->next), i++) {
-		resize(c, x, y, w - 2 * c->bw, /* remainder */ ((i + 1 == n)
-		       ? (wy + wh) - y : h) - 2 * c->bw, resizehints);
+		resize(c, x, y, NOBORDER(w), /* remainder */ ((i + 1 == n)
+		       ? NOBORDER(wy + wh) - y : h), resizehints);
 		if(h != wh)
 			y = c->y + c->h + 2 * c->bw;
 	}
@@ -1476,28 +1473,14 @@ togglefloating(const Arg *arg) {
 	arrange();
 }
 
-void
-togglelayout(const Arg *arg) {
-	if(arg && arg->v)
-		lt = (Layout *)arg->v;
-	else if(++lt == &layouts[LENGTH(layouts)])
-		lt = &layouts[0];
-	if(sel)
-		arrange();
-	else
-		drawbar();
-}
-
-void
-togglemax(const Arg *arg) {
-	ismax = !ismax;
-	arrange();
-}
-
 void
 toggletag(const Arg *arg) {
-	uint mask = sel->tags ^ (arg->ui & TAGMASK);
+	unsigned int mask;
 
+	if (!sel)
+		return;
+	
+	mask = sel->tags ^ (arg->ui & TAGMASK);
 	if(sel && mask) {
 		sel->tags = mask;
 		arrange();
@@ -1506,10 +1489,11 @@ toggletag(const Arg *arg) {
 
 void
 toggleview(const Arg *arg) {
-	uint mask = tagset[seltags] ^ (arg->ui & TAGMASK);
+	unsigned int mask = tagset[seltags] ^ (arg->ui & TAGMASK);
 
 	if(mask) {
 		tagset[seltags] = mask;
+		clearurgent();
 		arrange();
 	}
 }
@@ -1556,16 +1540,24 @@ updatebar(void) {
 void
 updategeom(void) {
 #ifdef XINERAMA
-	int i;
+	int n, i = 0;
 	XineramaScreenInfo *info = NULL;
 
 	/* window area geometry */
-	if(XineramaIsActive(dpy)) {
-		info = XineramaQueryScreens(dpy, &i);
-		wx = info[xidx].x_org;
-		wy = showbar && topbar ?  info[xidx].y_org + bh : info[xidx].y_org;
-		ww = info[xidx].width;
-		wh = showbar ? info[xidx].height - bh : info[xidx].height;
+	if(XineramaIsActive(dpy) && (info = XineramaQueryScreens(dpy, &n))) { 
+		if(n > 1) {
+			int di, x, y;
+			unsigned int dui;
+			Window dummy;
+			if(XQueryPointer(dpy, root, &dummy, &dummy, &x, &y, &di, &di, &dui))
+				for(i = 0; i < n; i++)
+					if(INRECT(x, y, info[i].x_org, info[i].y_org, info[i].width, info[i].height))
+						break;
+		}
+		wx = info[i].x_org;
+		wy = showbar && topbar ?  info[i].y_org + bh : info[i].y_org;
+		ww = info[i].width;
+		wh = showbar ? info[i].height - bh : info[i].height;
 		XFree(info);
 	}
 	else
@@ -1581,12 +1573,28 @@ updategeom(void) {
 	by = showbar ? (topbar ? wy - bh : wy + wh) : -bh;
 }
 
+void
+updatenumlockmask(void) {
+	unsigned int i, j;
+	XModifierKeymap *modmap;
+
+	numlockmask = 0;
+	modmap = XGetModifierMapping(dpy);
+	for(i = 0; i < 8; i++)
+		for(j = 0; j < modmap->max_keypermod; j++)
+			if(modmap->modifiermap[i * modmap->max_keypermod + j] == XKeysymToKeycode(dpy, XK_Num_Lock))
+				numlockmask = (1 << i);
+	XFreeModifiermap(modmap);
+}
+
 void
 updatesizehints(Client *c) {
 	long msize;
 	XSizeHints size;
 
-	XGetWMNormalHints(dpy, c->win, &size, &msize);
+	if(!XGetWMNormalHints(dpy, c->win, &size, &msize))
+		/* size is uninitialized, ensure that size.flags aren't used */
+		size.flags = PSize; 
 	if(size.flags & PBaseSize) {
 		c->basew = size.base_width;
 		c->baseh = size.base_height;
@@ -1626,13 +1634,13 @@ updatesizehints(Client *c) {
 	else
 		c->maxa = c->mina = 0.0;
 	c->isfixed = (c->maxw && c->minw && c->maxh && c->minh
-			&& c->maxw == c->minw && c->maxh == c->minh);
+	             && c->maxw == c->minw && c->maxh == c->minh);
 }
 
 void
 updatetitle(Client *c) {
 	if(!gettextprop(c->win, netatom[NetWMName], c->name, sizeof c->name))
-		gettextprop(c->win, wmatom[WMName], c->name, sizeof c->name);
+		gettextprop(c->win, XA_WM_NAME, c->name, sizeof c->name);
 }
 
 void
@@ -1640,19 +1648,25 @@ updatewmhints(Client *c) {
 	XWMHints *wmh;
 
 	if((wmh = XGetWMHints(dpy, c->win))) {
-		if(c == sel)
-			sel->isurgent = False;
+		if(ISVISIBLE(c) && wmh->flags & XUrgencyHint) {
+			wmh->flags &= ~XUrgencyHint;
+			XSetWMHints(dpy, c->win, wmh);
+		}
 		else
 			c->isurgent = (wmh->flags & XUrgencyHint) ? True : False;
+
 		XFree(wmh);
 	}
 }
 
 void
 view(const Arg *arg) {
+	if((arg->ui & TAGMASK) == tagset[seltags])
+		return;
 	seltags ^= 1; /* toggle sel tagset */
-	if(arg && (arg->ui & TAGMASK))
-		tagset[seltags] = arg->i & TAGMASK;
+	if(arg->ui & TAGMASK)
+		tagset[seltags] = arg->ui & TAGMASK;
+	clearurgent();
 	arrange();
 }
 
@@ -1693,7 +1707,7 @@ void
 zoom(const Arg *arg) {
 	Client *c = sel;
 
-	if(ismax || !lt->arrange || (sel && sel->isfloating))
+	if(!lt[sellt]->arrange || lt[sellt]->arrange == monocle || (sel && sel->isfloating))
 		return;
 	if(c == nexttiled(clients))
 		if(!c || !(c = nexttiled(c->next)))
@@ -1707,13 +1721,15 @@ zoom(const Arg *arg) {
 int
 main(int argc, char *argv[]) {
 	if(argc == 2 && !strcmp("-v", argv[1]))
-		eprint("dwm-"VERSION", © 2006-2008 dwm engineers, see LICENSE for details\n");
+		die("dwm-"VERSION", © 2006-2008 dwm engineers, see LICENSE for details\n");
 	else if(argc != 1)
-		eprint("usage: dwm [-v]\n");
+		die("usage: dwm [-v]\n");
+
+	if(!setlocale(LC_CTYPE, "") || !XSupportsLocale())
+		fprintf(stderr, "warning: no locale support\n");
 
-	setlocale(LC_CTYPE, "");
 	if(!(dpy = XOpenDisplay(0)))
-		eprint("dwm: cannot open display\n");
+		die("dwm: cannot open display\n");
 
 	checkotherwm();
 	setup();
